//
// Glideslope Button Control ... leveraging Enjo's MFDButtons helper library
//
// Purpose ... sets up the menus and supports the context page switching (main, config, deorbit menus)
//
// (c) Andrew Stokes (ADSWNJ) 2012
//
// All rights reserved
//

#include "MFDButtonPage.hpp"
#include "GSmfdButtons.hpp"
#include "Glideslope.hpp" 

GSmfdButtons::GSmfdButtons()
{
    // The menu descriptions of all buttons
    static const MFDBUTTONMENU mnu0[] =
    {
      {"Switch Mode", 0, 'M'},
      {"Config Screen", 0, 'C'},
      {"Switch Units", 0, 'U'},
      {"AutoPilot Toggle", 0, 'B'},
      {"Clear Track History", 0, 'T'},
      {"Save User Glideslope", 0, 'V'},

	    {"Switch HAC", 0, 'H'},
      {"Reset HAC & Zoom", 0, 'R'},
	    {"HAC Radius++", 0, '='},
      {"HAC Radius--", 0, '-'},
      {"Manual Zoom++", 0, '\''},
      {"Manual Zoom--", 0, ';'}
    };

    RegisterPage(mnu0, sizeof(mnu0) / sizeof(MFDBUTTONMENU));

    RegisterFunction("MOD", OAPI_KEY_M, &Glideslope::menuMode);
    RegisterFunction("CFG", OAPI_KEY_C, &Glideslope::menuConfig);
    RegisterFunction("UNT", OAPI_KEY_U, &Glideslope::menuUnits);
    RegisterFunction("AUT", OAPI_KEY_B, &Glideslope::menuAutopilot);
    RegisterFunction("CLR", OAPI_KEY_T, &Glideslope::menuClearGS);
    RegisterFunction("SAV", OAPI_KEY_V, &Glideslope::menuSaveGS);

  	RegisterFunction("HAC", OAPI_KEY_H, &Glideslope::menuHACgeometry);
    RegisterFunction("RST", OAPI_KEY_R, &Glideslope::menuResetHAC);
    RegisterFunction("HR+", OAPI_KEY_EQUALS, &Glideslope::menuHACRadUp);
    RegisterFunction("HR-", OAPI_KEY_MINUS, &Glideslope::menuHACRadDn);
    RegisterFunction("ZM+", OAPI_KEY_APOSTROPHE, &Glideslope::menuZoomUp);
    RegisterFunction("ZM-",  OAPI_KEY_SEMICOLON, &Glideslope::menuZoomDn);


    static const MFDBUTTONMENU mnu1[] =
    {
      {"Return", 0, 'K'},
      {"Previous Base", 0, 'A'},
      {"Next Base", 0, 'S'},
      {"Previous Rwy", 0, 'Q'},
      {"Next Rwy", 0, 'W'},
      {"Deorbit Screen", 0, 'O'},

      {"Previous Glideslope", 0, 'E'},
      {"Next Glideslope", 0, 'R'},
	    {"Diagnostic Screen", 0, 'D'},
    };

    RegisterPage(mnu1, sizeof(mnu1) / sizeof(MFDBUTTONMENU));

    RegisterFunction("OK", OAPI_KEY_K, &Glideslope::menuOK);
    RegisterFunction("PB", OAPI_KEY_A, &Glideslope::menuPrevBase);
    RegisterFunction("NB", OAPI_KEY_S, &Glideslope::menuNextBase);
    RegisterFunction("PR", OAPI_KEY_Q, &Glideslope::menuPrevRwy);
    RegisterFunction("NR", OAPI_KEY_W, &Glideslope::menuNextRwy);
    RegisterFunction("DEO",OAPI_KEY_O, &Glideslope::menuDeorbitSel);

    RegisterFunction("PG", OAPI_KEY_E, &Glideslope::menuPrevGS);
    RegisterFunction("NG", OAPI_KEY_R, &Glideslope::menuNextGS);
    RegisterFunction("DIA", OAPI_KEY_D, &Glideslope::menuDiagSel);

    static const MFDBUTTONMENU mnu2[] =
    {
      {"Return", 0, 'K'},
    };

    RegisterPage(mnu2, sizeof(mnu2) / sizeof(MFDBUTTONMENU));

    RegisterFunction("OK", OAPI_KEY_K, &Glideslope::menuOK);

    static const MFDBUTTONMENU mnu3[] =
    {
      {"Switch Mode", 0, 'M'},
      {"Deorbit HP++", 0, '='},
      {"Deorbit HP+", 0, '\''},
      {"Deorbit HP-", 0, ';'},
      {"Deorbit HP--", 0, '-'},
    };

    RegisterPage(mnu3, sizeof(mnu3) / sizeof(MFDBUTTONMENU));

    RegisterFunction("MOD",  OAPI_KEY_M, &Glideslope::menuMode);
    RegisterFunction("H++",  OAPI_KEY_EQUALS, &Glideslope::menuDeorHPup);
    RegisterFunction("H+",   OAPI_KEY_APOSTROPHE, &Glideslope::menuDeorHPsmUp);
    RegisterFunction("H-",   OAPI_KEY_SEMICOLON, &Glideslope::menuDeorHPsmDn);
    RegisterFunction("H--",  OAPI_KEY_MINUS, &Glideslope::menuDeorHPdn);

    return;
}

bool GSmfdButtons::SearchForKeysInOtherPages() const
{
    return false;
}
