//
// Glideslope Core Functions
//
// Purpose ... persistent class supporting the core calculations for GS2
// (The Glideslope class is very ephemeral ... it gets destroyed and reinstantiated each time you change view, etc, so persist the core into this class)
//
//
// (c) Andrew Stokes (ADSWNJ) 2012
//
// All rights reserved
//

#include "windows.h"
#include "orbitersdk.h"
#include "DisplayEngUnitFunctions.h"
#include "ParseFunctions.h"
#include "GSMiscFunctions.hpp"
#include "MFDButtonPage.hpp"
#include "GSmfdButtons.hpp"

#ifndef _GS_CORE_CLASS
#define _GS_CORE_CLASS

class GlideslopeCore {
  public:
  VESSEL *vessel;
  //Tracking variables
  int TrackPtr;															// Current track point
  double TrackT[10000];													// Simulation time
  double TrackD[10000];													// Range (distance)
  double TrackV[10000];													// Velocity (TAS)
  double TrackA[10000];													// Altitude
  double TrackAlpha[10000];												// AoA (Alpha)
  double TrackRefAlpha[10000];											
  double TrackVspd[10000];
  double TrackBeta[10000];
  double TrackGamma[10000];
  double Trackvreldot[10000];
  double TrackRefvreldot[10000];
  VECTOR3 xyz_track[10000];
  void calcRunway();
  void calcRange(); //Range to runway touchdown point via waypoints and HAC as necessary
  void MinorCycle(double simt);
  void MajorCycle(double simt);
  void reset(VESSEL *v);
  void steer(double YawCmd, bool YawActive,double PitchCmd, bool PitchActive,double RollCmd, bool RollActive);
  void saveUserGS(bool units);

  bool isInit;
  int coreMFDcount;
  static const int CORE_MFD_LIMIT = 10;
  int coreMFDint[CORE_MFD_LIMIT];
  int coreMode[CORE_MFD_LIMIT];
  bool coreUnit[CORE_MFD_LIMIT];
  bool coreManZoomToggle[CORE_MFD_LIMIT];
  double coreZoom[CORE_MFD_LIMIT];
  GSmfdButtons coreButtons[CORE_MFD_LIMIT];

  static const int RUNWAY_LIMIT=128;
  static const int BASE_LIMIT=64;
  static const int GSFILE_LIMIT=64;
  static const int GLIDESLOPE_POINTS_LIMIT=256;
  static const int BETA_GUIDE_OFF_ALT=20000; //meters, ~60kft
  static const int EI_ALT=120000; // used to be 121920



  bool steerActive;
  bool diagOn;
  bool deorOn;
  int runway;
  int GSfile;

  int RunwayCount;
  int BaseCount;
  int GSfileCount;
  double RunwayData[RUNWAY_LIMIT][9];
  int RunwayBase[RUNWAY_LIMIT];
  char RunwayName[RUNWAY_LIMIT][12];
  char BaseName[BASE_LIMIT][32];
  char GSfileName[GSFILE_LIMIT][32];
  char GSname[GSFILE_LIMIT][32];
  double BaseLL[BASE_LIMIT][2];
  int ParamsModeL, ParamsModeR, ParamsModeX, ParamsRwy;
  bool ParamsUnits; 
  int RefSlopePts;
  double RefSlope[GLIDESLOPE_POINTS_LIMIT][6];
  int maxMod;
  int hacRadius;
  int finalDist;
  double re;													//Radius of the focus planet (usually Earth)



  //HAC runway constants
  //All vectors are of the form xxx_yyy_zzz where
  //xxx is llr latitude, longitude, radius (no zzz)
  //       xyz ECF cartesian (no zzz)
  //       nre north, radial, east relative to zzz
  //       ar  relative to zzz
  //       M   xyz-to-nre matrix for this point (no zzz)
  //yyy is the name of this point
  //zzz is the name of the reference point

  //Base coordinates
  VECTOR3 llr_base,xyz_base;									//Coordinates of our base rwy in LLR (lat,lon,radius) and XYZ format (ECEF Cartesian Coords)
  MATRIX3 M_base;												//Mapping matrix from NRE vectors to XYZ coords (e.f. for near/far ends of base rwy)
  //Runway near threshold points
  VECTOR3 xyz_near, nre_near_base, nre_near_far, nre_near_sc;	// XYZ absolue, then NRE relative to base, relative to far end and relative to spacecraft
  AZ_RANGE ar_near_far;
  //Runway far threshold points
  VECTOR3 xyz_far, nre_far_base, nre_far_sc;
  MATRIX3 M_far;
  //PAPI point
  double disp_papi;
  VECTOR3 nre_papi_far,xyz_papi;
  AZ_RANGE ar_papi_far;
  //Touchdown point
  double disp_tdp;
  VECTOR3 nre_tdp_far,xyz_tdp;
  AZ_RANGE ar_tdp_far;
  //WP2 point
  double disp_wp2,ra_wp2_far;
  VECTOR3 nre_wp2_far,xyz_wp2,nre_wp2_hac,nre_wp2_sc;
  AZ_RANGE ar_wp2_far,ar_wp2_hac;
  //Runway constants
  double rwy_length;
  VECTOR3 rwy_pole;
  //HAC
  double base_radius_hac, radius_hac,ra_hac,ra_hac_far,daz_hac_far;
  VECTOR3 nre_hac_far,xyz_hac;
  AZ_RANGE ar_hac_far;
  MATRIX3 M_hac;
  int hac_sign;													// 1 for Anti-Clockwise, -1 for clockwise
  int hac_open;													// 0 for Closed (do a lap before arming WP2), 1 for Open (less than a lap)
  int RangeSeg;													// -1 whilst outside ref glideslope
  //HAC spacecraft constants
  //Spacecraft
  VECTOR3 llr_sc,xyz_sc,nre_sc_far,nre_sc_hac,nre_hac_sc;
  AZ_RANGE ar_sc_hac,ar_hac_sc,ar_sc_far;
  MATRIX3 M_sc;
  //WP1
  AZ_RANGE ar_wp1_sc,ar_wp1_hac;
  VECTOR3 nre_wp1_sc,nre_wp1_far,nre_wp1_hac,xyz_wp1;
  double crossrange,crossrange_angle,az_tgt_sc,range_papi_sc,range;
  int leg, maxleg;
  double range_wp1_sc;
  double arc_range, hac_arc;
  bool wp2Armed, finalArmed; 

  OBJHANDLE Planet;
  double GtimesM1;
  double M2;
  VECTOR3 AirspeedVec,InertialVel;
  double Lat,Lon,Altitude,vspd,Groundspeed,Bearing,Mach,Airspeed,AirspeedRate,RefAirspeedRate,DynPres,HeatFlux,InertialSpd,DelAz,alpha,beta,gamma;
  double lastSimT,lastTE,lastRefTE,lastRefVspd,lastVspd,lastBearing,lastAirspeed,lastRefAltitude,lastRefAirspeed,lastAltitude,nextMajorCycle;
  double PE,KE,TE,RefPE,RefKE,RefTE,RefVspdRate,RefTERate,BearingRate,VspdRate,TERate;
  double RefVspd,Refalpha,RefAirspeed,RefAltitude;
  double lastAlpha,lastBeta,lastGamma,lastMinorSimT;
  bool hasRefSlope;

  //Trim control
  double TrimPos,TrimSpd,TrimLastSimT,TrimGain;

  //Deorbit calculator
  double DeorbitHp;
};
#endif // _GS_CORE_CLASS